#
# File:    PdbxSFMapCoefficients.py
##
"""  Classes to aid in conversion and manipulation of SF map coefficient files from the validation package.
"""

__docformat__ = "restructuredtext en"
__author__ = "Ezra Peisach"
__email__ = "ezra.peisach@rcsb.org"
__license__ = "Apache 2.0"

import logging
import copy
import tempfile
import shutil
import os

from mmcif.io.IoAdapterCore import IoAdapterCore
from mmcif.api.PdbxContainers import DataContainer
from wwpdb.utils.dp.RcsbDpUtility import RcsbDpUtility
from wwpdb.utils.config.ConfigInfo import getSiteId

logger = logging.getLogger(__name__)


class PdbxSFMapCoefficients(object):
    def __init__(self, siteid=None, tmppath="/tmp", cleanup=True):
        self.__sf = None
        self.__siteid = getSiteId(siteid)
        self.__cleanup = cleanup
        self.__tmppath = tmppath

    def read_mmcif_sf(self, pathin):
        """Reads PDBx/mmCIF structure factor file with map coefficients

        Return True on success, otherwise False
        """

        logger.debug("Starting read %s", pathin)
        try:
            io = IoAdapterCore()
            self.__sf = io.readFile(pathin)
            return True
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.__sf = None
            return False

    def has_map_coeff(self):
        """Returns True if read in SF file has map coefficients, else returns False"""
        if self.__sf is None:
            return False

        if len(self.__sf) == 0:
            return False

        # Check first block
        b0 = self.__sf[0]

        c0 = b0.getObj("refln")
        if c0 is None:
            return False

        alist = c0.getAttributeList()
        for att in ["index_h", "index_k", "index_l", "fom", "pdbx_DELFWT", "pdbx_DELPHWT", "pdbx_FWT", "pdbx_PHWT"]:
            if att not in alist:
                logger.debug("Missing %s from sf file", att)
                return False

        return True

    def read_mtz_sf(self, pathin):
        """Reads MTZ structure factor file

        Return True on success, otherwise False
        """

        logger.debug("Starting mtz read %s", pathin)

        suffix = "-dir"
        prefix = "rcsb-"
        if self.__tmppath is not None and os.path.isdir(self.__tmppath):
            workpath = tempfile.mkdtemp(suffix, prefix, self.__tmppath)
        else:
            workpath = tempfile.mkdtemp(suffix, prefix)

        diagfn = os.path.join(workpath, "sf-convert-diags.cif")
        ciffn = os.path.join(workpath, "sf-convert-datafile.cif")
        dmpfn = os.path.join(workpath, "sf-convert-mtzdmp.log")
        logfn = os.path.join(workpath, "sf-convert.log")

        dp = RcsbDpUtility(siteId=self.__siteid, tmpPath=self.__tmppath)
        dp.imp(pathin)
        dp.op("annot-sf-convert")
        dp.expLog(logfn)
        dp.expList(dstPathList=[ciffn, diagfn, dmpfn])
        if os.path.exists(ciffn):
            ret = self.read_mmcif_sf(ciffn)
        else:
            ret = False

        if self.__cleanup:
            dp.cleanup()
            shutil.rmtree(workpath, ignore_errors=True)
        return ret

    def write_mmcif_coef(self, fopathout, twofopathout, entry_id="xxxx"):
        """Writes out two structure factor files with only fo-fc or 2fo-fc coefficients

        Output files are dictionary compliant

        entry.id will be set to entry_id
        """
        ret1 = self.__write_mmcif(fopathout, "fo", entry_id)
        ret2 = self.__write_mmcif(twofopathout, "2fo", entry_id)
        return ret1 and ret2

    def __write_mmcif(self, pathout, coef, entry_id):
        """Writes out the specific map coefficients"""

        # Categories that will not be copied
        _striplist = ["audit", "diffrn_radiation_wavelength", "exptl_crystal", "reflns_scale"]

        # refln attributes to keep
        _keepattr = ["index_h", "index_k", "index_l", "fom"]
        if coef == "fo":
            _keepattr.extend(["pdbx_DELFWT", "pdbx_DELPHWT"])
        else:
            _keepattr.extend(["pdbx_FWT", "pdbx_PHWT"])

        # Datablockname
        blkname = "{}{}".format(entry_id, coef)
        new_cont = DataContainer(blkname)

        # Only care about first block
        blockin = self.__sf[0]

        for objname in blockin.getObjNameList():
            if objname in _striplist:
                continue

            myobj = blockin.getObj(objname)

            # Make a copy of the original - as likely will need to modify
            modobj = copy.deepcopy(myobj)
            if objname == "entry":
                modobj.setValue(entry_id, "id", 0)
            if objname in ["cell", "symmetry"]:
                modobj.setValue(entry_id, "entry_id", 0)
            if objname == "refln":
                # Remove all but what we want
                # Make a copy to ensure not messed with during operation
                for attr in list(modobj.getAttributeList()):
                    if attr not in _keepattr:
                        modobj.removeAttribute(attr)

            new_cont.append(modobj)

        # new_cont.printIt()
        io = IoAdapterCore()
        # Write out a single block
        ret = io.writeFile(pathout, [new_cont])
        return ret



import os
import sys
import time
import logging
import subprocess
import argparse
from wwpdb.utils.config.ConfigInfo import ConfigInfo, getSiteId

logger = logging.getLogger()


class RunRemote:
    def __init__(self, command, job_name,
                 log_dir,
                 run_dir=None,
                 timeout=5600,
                 memory_limit=0,
                 number_of_processors=1,
                 add_site_config=False,
                 add_site_config_database=False):
        self.command = self.escape_substitution(command)
        if timeout:
            self.timeout = timeout
        else:
            self.timeout = 5600
        self.memory_limit = memory_limit
        self.number_of_processors = number_of_processors
        self.job_name = job_name
        self.log_dir = log_dir
        self.run_dir = run_dir
        self.memory_used = 0
        self.memory_unit = "MB"
        self.bsub_exit_status = 0
        self.siteId = getSiteId()
        self.cI = ConfigInfo(self.siteId)
        self.bsub_source_command = self.cI.get("BSUB_SOURCE")
        self.bsub_run_command = self.cI.get("BSUB_COMMAND")
        self.pdbe_cluster_queue = self.cI.get("PDBE_CLUSTER_QUEUE")
        self.pdbe_memory_limit = 100000
        self.bsub_login_node = self.cI.get("BSUB_LOGIN_NODE")
        self.bsub_timeout = self.cI.get("BSUB_TIMEOUT")
        self.bsub_retry_delay = self.cI.get("BSUB_RETRY_DELAY", 4)
        self.command_prefix = self.cI.get('REMOTE_COMMAND_PREFIX')
        self.bsub_log_file = os.path.join(self.log_dir, self.job_name + ".log")
        self.bsub_out_file = os.path.join(self.log_dir, self.job_name + ".out")
        self.add_site_config = add_site_config
        self.add_site_config_database = add_site_config_database

        self.out = None
        self.err = None

    def escape_substitution(self, command):
        """
        Escapes dollars, stops variables being interpretted early when passed to bsub.
        """
        command = command.replace("$", "\$")  # noqa: W605 pylint: disable=anomalous-backslash-in-string
        return command

    def write_run_script(self):
        self.command = self.escape_substitution(self.command)
        if self.run_dir:
            shell_script = os.path.join(self.run_dir, 'run_{}.sh'.format(self.job_name))
            with open(shell_script, 'w') as out_file:
                out_file.write(self.command)
            os.chmod(shell_script, 0o775)
            self.command = shell_script

    def run(self):
        rc = 1

        self.write_run_script()

        if self.add_site_config_database:
            self.pre_pend_sourcing_site_config(database=True)
        if self.add_site_config:
            self.pre_pend_sourcing_site_config()
        if self.command_prefix:
            self.prefix_command()

        if self.bsub_run_command:
            bsub_try = 1
            rc, self.out, self.err = self.run_bsub()
            while self.bsub_exit_status != 0:
                if self.memory_used:
                    try:
                        if self.memory_used > self.memory_limit:
                            self.memory_limit = int(self.memory_used)
                    except:  # noqa: E722 pylint: disable=bare-except
                        pass

                if self.memory_limit >= 100000:
                    self.memory_limit = self.memory_limit + 40000
                elif self.memory_limit >= 20000:
                    self.memory_limit = self.memory_limit + 30000
                else:
                    self.memory_limit = self.memory_limit + 10000
                bsub_try += 1
                logging.info("try {}, memory {}".format(bsub_try, self.memory_limit))  # pylint: disable=logging-format-interpolation
                rc, self.out, self.err = self.run_bsub()

        if rc != 0:
            logging.error("return code: {}".format(rc))  # pylint: disable=logging-format-interpolation
            logging.error("out: {}".format(self.out))  # pylint: disable=logging-format-interpolation
            logging.error("error: {}".format(self.err))  # pylint: disable=logging-format-interpolation
        else:
            logging.info("worked")

        return rc

    @staticmethod
    def check_timing(t1, t2):
        t = t2 - t1
        human_time = []
        if t > 3600:
            human_time.append("%.2f hours" % (t / 3600))
        elif t > 60:
            human_time.append("%.2f minutes" % (t / 60))
        else:
            human_time.append("%.2f seconds" % t)

        abs_time = "TIMING, %.2f, minutes" % (t / 60)

        human_time.append(abs_time)

        return human_time

    @staticmethod
    def touch(fname):
        if os.path.exists(fname):
            os.utime(fname, None)
        else:
            open(fname, "a").close()

    def get_site_config_command(self, suffix=""):
        site_config_path = self.cI.get("TOP_WWPDB_SITE_CONFIG_DIR")
        site_loc = self.cI.get("WWPDB_SITE_LOC")
        site_config_command = ". {}/init/env.sh --siteid {} --location {} {} > /dev/null;".format(site_config_path, self.siteId, site_loc, suffix)
        return site_config_command

    def pre_pend_sourcing_site_config(self, database=False):

        self.command = "{}; {}".format(self.get_site_config_command(), self.command)
        if database:
            self.command = "{}; {}".format(self.get_site_config_command(suffix="--database"), self.command)

    def prefix_command(self):
        if self.command_prefix:
            self.command = "{} {}".format(self.command_prefix, self.command)

    def check_bsub_finished(self):
        # pause to allow system to write out bsub out file.
        time.sleep(5)
        if not os.path.exists(self.bsub_out_file):
            retries = 0
            logging.info("bsub out file not present - waiting for bsub to finish")
            max_num_of_retries = int(self.bsub_timeout) / int(self.bsub_retry_delay)
            while retries < max_num_of_retries:
                if os.path.exists(self.bsub_out_file):
                    logging.info("found bsub out file")
                    break
                else:
                    retries += 1
                    logging.info("try {} of {}, wait for {} seconds".format(retries, max_num_of_retries, self.bsub_retry_delay))  # pylint: disable=logging-format-interpolation
                    time.sleep(int(self.bsub_retry_delay))

    def run_command(self, command, log_file=None, new_env=None):
        # command_list = shlex.split(command)
        logging.info("Starting: %s", self.job_name)
        logging.info(command)
        if log_file:
            logging.info("logging to: {}".format(log_file))  # pylint: disable=logging-format-interpolation
            if not os.path.exists(os.path.dirname(log_file)):
                os.makedirs(os.path.dirname(log_file))
        t1 = os.times()[4]
        # child = subprocess.Popen(command_list)
        if new_env:
            child = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=new_env)
        else:
            child = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = child.communicate()
        rc = child.returncode
        if rc != 0:
            logging.error("Exit status: %s - process failed: %s", rc, self.job_name)
        else:
            logging.info("process worked: %s", self.job_name)

        if log_file:
            with open(log_file, "wb") as lf:
                if out:
                    lf.write(out)
                    # logging.info(out)
                if err:
                    lf.write(err)
                    logging.error(err)

        t2 = os.times()[4]
        ht = self.check_timing(t1, t2)
        # logging.info("Timing: %s took %s" %(name, ht[0]))
        logging.info("Finished: %s, %s", self.job_name, ht[1])

        return rc, out, err

    def launch_bsub(self):

        if os.path.exists(self.bsub_out_file):
            os.remove(self.bsub_out_file)

        bsub_command = list()
        if self.bsub_login_node:
            bsub_command.append("ssh {} '".format(self.bsub_login_node))
        if self.bsub_source_command:
            bsub_command.append("{};".format(self.bsub_source_command))
        bsub_command.append(self.bsub_run_command)
        bsub_command.append("-J {}".format(self.job_name))
        bsub_command.append("-oo {}".format(self.bsub_log_file))
        bsub_command.append("-eo {}/{}_error.log".format(self.log_dir, self.job_name))
        bsub_command.append('-Ep "touch {}"'.format(self.bsub_out_file))
        if self.pdbe_memory_limit and self.memory_limit > self.pdbe_memory_limit:
            bsub_command.append("-P {}".format("bigmem"))

        bsub_command.append("-q {}".format(self.pdbe_cluster_queue))
        if "LSB_JOBGROUP" in os.environ and os.environ["LSB_JOBGROUP"]:
            bsub_command.append("-g {}".format(os.environ["LSB_JOBGROUP"]))

        bsub_command.append("-n {}".format(self.number_of_processors))
        bsub_command.append("-W {}".format(self.timeout))
        if self.memory_limit:
            bsub_command.append('-R "rusage[mem={0}]" -M {0}'.format(self.memory_limit))
        bsub_command.append('-K "{}"'.format(self.command))
        # bsub_command.append('"{}"'.format(self.command))
        if self.bsub_login_node:
            bsub_command.append("'")

        command_string = " ".join(bsub_command)

        # if self.run_dir:
        #     shell_script = os.path.join(self.run_dir, 'bsub_command_{}.sh'.format(self.job_name))
        #     with open(shell_script, 'w') as out_file:
        #         out_file.write(command_string)
        #     os.chmod(shell_script, 0o775)

        rc, out, err = self.run_command(command=command_string)

        return rc, out, err

    def launch_bsub_wait_process(self):
        bsub_command = list()
        bsub_command.append("{};".format(self.bsub_source_command))
        bsub_command.append(self.bsub_run_command)
        bsub_command.append('-J "end_{}"'.format(self.job_name))
        bsub_command.append('-w "ended({})"'.format(self.job_name))
        bsub_command.append('-oo "{}/{}_wait.log"'.format(self.log_dir, self.job_name))
        bsub_command.append('-eo "{}/{}_wait_error.log"'.format(self.log_dir, self.job_name))
        bsub_command.append("-q {}".format(self.pdbe_cluster_queue))
        bsub_command.append('-K "uname -a; date"')
        command_string = " ".join(bsub_command)
        rc, out, err = self.run_command(command=command_string)

        return rc, out, err

    def parse_bsub_log(self):
        self.bsub_exit_status = 0
        self.memory_used = 0
        self.memory_unit = "MB"
        if os.path.exists(self.bsub_log_file):
            with open(self.bsub_log_file, "r") as log_file:
                for l in log_file:  # noqa: E741
                    if "Max Memory :" in l:
                        try:
                            memory_used = l.split(":")[-1].strip()
                            self.memory_unit = memory_used.split(" ")[1]
                            self.memory_used = int(memory_used.split(" ")[0])
                        except Exception as e:
                            logging.error(e)

                    if "TERM_MEMLIMIT" in l:
                        self.bsub_exit_status = 1
        if self.memory_unit == "GB":
            self.memory_unit = "MB"
            self.memory_used = self.memory_used * 1024
        elif self.memory_unit == "KB":
            self.memory_unit = "MB"
            self.memory_used = int(self.memory_used / 1024)
        logging.info("memory used: {} {}".format(self.memory_used, self.memory_unit))  # pylint: disable=logging-format-interpolation
        logging.info("bsub exit status: {}".format(self.bsub_exit_status))  # pylint: disable=logging-format-interpolation

    def run_bsub(self):

        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir)

        temp_file = os.path.join(self.log_dir, "bsub_temp_file.out")

        # if get non ok exit status from bsub then wait 30 seconds and try again.
        i = 0
        rc, out, err = 0, None, None

        # error codes
        # 0 everything is ok
        # 143 memory limit reached
        # 159/153 file too large - need additional resources, trying again wont help
        # 255 is ssh connection dropped so task is still ongoing - this is also when lsf is not ready
        allowed_codes = (0, 143, 153, 159)

        while i < 10:
            rc, out, err = self.launch_bsub()
            if rc in allowed_codes:
                break
            delay_time = i * 2
            logging.info("bsub return code of {}. Waiting for {}".format(rc, delay_time))  # pylint: disable=logging-format-interpolation
            time.sleep(delay_time)
            i += 1

        if rc not in allowed_codes:
            return rc, out, err

        self.touch(temp_file)
        self.check_bsub_finished()
        self.parse_bsub_log()

        return rc, out, err


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--debug", help="debugging", action="store_const", dest="loglevel", const=logging.DEBUG, default=logging.INFO)
    parser.add_argument("--command", help="command to run", type=str, required=True)
    parser.add_argument("--job_name", help="name for the job", type=str, required=True)
    parser.add_argument("--log_dir", help="directory to store log file in", type=str, required=True)
    parser.add_argument("--run_dir", help="directory to run", type=str)
    parser.add_argument("--memory_limit", help="starting memory limit", type=int, default=0)
    parser.add_argument("--num_processors", help="number of processors", type=int, default=1)
    parser.add_argument("--add_site_config", help="add site config to command", action="store_true")
    parser.add_argument("--add_site_config_with_database", help="add site config with database to command", action="store_true")

    args = parser.parse_args()
    logger.setLevel(args.loglevel)

    run_remote = RunRemote(
        command=args.command,
        job_name=args.job_name,
        log_dir=args.log_dir,
        run_dir=args.run_dir,
        memory_limit=args.memory_limit,
        number_of_processors=args.num_processors,
        add_site_config=args.add_site_config,
        add_site_config_database=args.add_site_config_with_database,
    )

    ret = run_remote.run()

    if ret != 0:
        message = "{} failed".format(args.job_name)
        annotation_email = run_remote.cI.get("ANNOTATION_EMAIL")
        if annotation_email:
            mcommand = 'mail -s "{} failed" {}'.format(args.job_name, annotation_email)
            run_remote.run_command(command=mcommand)

    sys.exit(ret)



# File:    VaildationWrapper.py
#
# Updates:
# 30-Jun-2020 zf  Add image tar file output after svg file for "annot-wwpdb-validate-all", "annot-wwpdb-validate-all-v2" & "annot-wwpdb-validate-all-sf"
##
""" Class to act as a front end to validation calculations and repackaging.
    Access pattern mirrors RscbDpUtility.

    Provides an op annot-wwpdb-validate-all-sf which provides SF coefficients by intercepting output ops
"""

__docformat__ = "restructuredtext en"
__author__ = "Ezra Peisach"
__email__ = "ezra.peisach@rcsb.org"
__license__ = "Apache 2.0"

import logging
import os
import sys

from mmcif.io.IoAdapterCore import IoAdapterCore
from wwpdb.utils.dp.RcsbDpUtility import RcsbDpUtility
from wwpdb.utils.dp.PdbxSFMapCoefficients import PdbxSFMapCoefficients
from wwpdb.io.file.DataFile import DataFile

logger = logging.getLogger(__name__)


class ValidationWrapper(RcsbDpUtility):
    def __init__(self, tmpPath="/scratch", siteId="DEV", verbose=False, log=sys.stderr):
        logger.debug("Starting")
        super(ValidationWrapper, self).__init__(tmpPath=tmpPath, siteId=siteId, verbose=verbose, log=log)
        self.__op = None
        self._tmppath = tmpPath
        self.__siteId = siteId
        self.__modelfile = None

    def __getPDBId(self):
        """Returns the PDB accession code in model file or None"""
        if not self.__modelfile:
            return None
        io = IoAdapterCore()
        cont = io.readFile(self.__modelfile, selectList=["database_2"])
        if cont:
            block = cont[0]
            catObj = block.getObj("database_2")
            if catObj:
                vals = catObj.selectValuesWhere("database_code", "PDB", "database_id")
                if len(vals) > 0 and vals[0] and vals[0] and len(vals[0]) > 0 and vals[0] not in [".", "?"]:
                    return vals[0]

        return None

    def imp(self, srcPath=None):
        self.__modelfile = srcPath
        super(ValidationWrapper, self).imp(srcPath)

    def op(self, op):
        logger.info("Starting op %s", op)
        if op not in ["annot-wwpdb-validate-all", "annot-wwpdb-validate-all-v2", "annot-wwpdb-validate-all-sf"]:
            logger.error("Operation not known %s", op)
            return False
        self.__op = op

        if op == "annot-wwpdb-validate-all-sf":
            op = "annot-wwpdb-validate-all-v2"
        return super(ValidationWrapper, self).op(op)

    def expList(self, dstPathList=None):
        if dstPathList is None:
            dstPathList = []

        if self.__op in ["annot-wwpdb-validate-all", "annot-wwpdb-validate-all-v2"]:
            return super(ValidationWrapper, self).expList(dstPathList)

        logger.debug("Intercepting expList")
        # Op should be annot-wwpdb-validate-all-sf - so we will "pull" and interpret

        # Handle the first arguments
        # ofpdf,ofxml,offullpdf,ofpng,ofsvg,ofmtz

        basedst = dstPathList[0:7]
        outfosf = dstPathList[7]
        out2fosf = dstPathList[8]
        mtzfile = os.path.join(self.getWorkingDir(), "mapcoef.mtz")
        basedst.append(mtzfile)
        ret = super(ValidationWrapper, self).expList(basedst)

        logger.debug("expList ret is %s", ret)

        # Need to "produce files" for last arguments if possible
        if not os.path.exists(mtzfile):
            return False

        pdbid = self.__getPDBId()
        logger.info("pdbID %s", pdbid)
        if not pdbid:
            pdbid = "xxxx"

        psm = PdbxSFMapCoefficients(siteid=self.__siteId, tmppath=self._tmppath)
        ret = psm.read_mtz_sf(mtzfile)
        logger.debug("read_mtz_sf ret %s", ret)

        if ret:
            # Ensure map coefficients produced in conversion - returns True if present
            ret = psm.has_map_coeff()
            logger.debug("Check for map coeffcients returns %s", ret)
            scrpath = self.getWorkingDir()
            fotemp = os.path.join(scrpath, "fotemp.cif")
            twofotemp = os.path.join(scrpath, "twofotemp.cif")

            if ret:
                ret = psm.write_mmcif_coef(fopathout=fotemp, twofopathout=twofotemp, entry_id=pdbid.lower())
                df = DataFile(fotemp)
                if df.srcFileExists():
                    df.copy(outfosf)
                df = DataFile(twofotemp)
                if df.srcFileExists():
                    df.copy(out2fosf)

        logger.debug("Returning %s", ret)
        return ret



##
# File:    PdbxStripCategory.py
# Author:  jdw
# Date:    17-Dec-2012
# Version: 0.001
#
# Updated:
#
##
"""  Remove selected categories from the first data container and
     write the result.
"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import logging
import sys

from mmcif.api.PdbxContainers import DataContainer
from mmcif.io.PdbxReader import PdbxReader
from mmcif.io.PdbxWriter import PdbxWriter

logger = logging.getLogger(__name__)


class PdbxStripCategory(object):
    def __init__(self, verbose=False, log=sys.stderr):  # pylint: disable=unused-argument
        pass

    def strip(self, inpPath, outPath, stripList=None):
        """Strip categories from inpPath and write to outPath"""
        if stripList is None:
            stripList = []

        try:
            myDataList = []
            with open(inpPath, "r") as ifh:
                pRd = PdbxReader(ifh)
                pRd.read(myDataList)
            #
            myBlock = myDataList[0]
            myName = myBlock.getName()
            newContainer = DataContainer(myName)

            for objName in myBlock.getObjNameList():
                myObj = myBlock.getObj(objName)
                if myObj.getName() not in stripList:
                    newContainer.append(myObj)
            #
            with open(outPath, "w") as ofh:
                pWr = PdbxWriter(ofh)
                pWr.setPreferSingleQuotes()
                pWr.write([newContainer])

            return True
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            return False


def _maintest():
    stripList = [
        "pdbx_coord",
        # 'pdbx_entity_nonpoly',
        # 'pdbx_missing_residue_list',
        "pdbx_nonstandard_list",
        "pdbx_protein_info",
        "pdbx_solvent_info",
        "pdbx_struct_sheet_hbond",
        "pdbx_unobs_or_zero_occ_residues",
        "pdbx_validate_torsion",
        "struct_biol_gen",
        "struct_conf",
        "struct_conf_type",
        "struct_mon_prot_cis",
        "struct_sheet",
        "struct_sheet_order",
        "struct_sheet_range",
    ]

    strp = PdbxStripCategory(verbose=True, log=sys.stderr)
    strp.strip("test-in.cif", "test-out.cif", stripList)


if __name__ == "__main__":
    _maintest()



##
# File:    DataFileAdapter.py
# Date:    16-July-2012
#
# Updates:
#
#  4-Jan-2013 -jdw  add filters to remove derived categories after format conversion
# 12-Feb-2013  jdw  add option flags for stipping derived categories --
# 25-Feb-2013  jdw  add cif2pdb translation.
#  6-Aug-2013  jdw  add rcsb2PdbxWithPdbIdAlt()
# 11-Oct-2013  jdw  add option to strip  additional entity categories --
#  4-Dec-2013  jdw  check for duplicate source and destination paths.
# 24-Dec-2013  jdw  add pdbx2Assemblies(idCode,inpFilePath,outPath='.',idxFilePath=None)
# 24-Dec-2013  jdw  add cif2pdbx()
# 16-Jan-2014  jdw  Update method for cif2pdbx()
# 25-Sep-2014  jdw  add pdbx2nmrstar()
##
"""
Encapsulate data model format type conversions.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

import sys
import os.path
import shutil
import traceback
from wwpdb.utils.dp.RcsbDpUtility import RcsbDpUtility


class DataFileAdapter(object):
    """Convenience methods to manage data model and structure factor format conversions."""

    def __init__(self, reqObj, verbose=False, log=sys.stderr):
        self.__reqObj = reqObj
        self.__verbose = verbose
        self.__lfh = log
        self.__debug = True
        self.__siteId = self.__reqObj.getValue("WWPDB_SITE_ID")
        self.__sObj = self.__reqObj.getSessionObj()
        # self.__sessionId = self.__sObj.getId()
        self.__sessionPath = self.__sObj.getPath()

    def pdbx2nmrstar(self, inpPath, outPath, pdbId=None):
        """PDBx to NMRSTAR"""
        try:
            dp = RcsbDpUtility(tmpPath=self.__sessionPath, siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
            dp.imp(inpPath)
            if pdbId is not None:
                dp.addInput(name="pdb_id", value=pdbId, type="param")
            dp.op("annot-pdbx2nmrstar")
            logPath = os.path.join(self.__sessionPath, "annot-pdbx2nmrstar.log")
            dp.expLog(logPath)
            dp.exp(outPath)
            if not self.__debug:
                dp.cleanup()
        except:  # noqa: E722 pylint: disable=bare-except
            traceback.print_exc(file=self.__lfh)
            return False
        #
        return True

    def rcsb2Pdbx(self, inpPath, outPath, stripFlag=False, stripEntityFlag=False):
        """RCSB CIF -> PDBx conversion  (Using the smaller application in the annotation package)"""
        try:
            dp = RcsbDpUtility(tmpPath=self.__sessionPath, siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
            dp.imp(inpPath)

            if stripFlag:
                if stripEntityFlag:
                    dp.op("annot-rcsb2pdbx-strip-plus-entity")
                else:
                    dp.op("annot-rcsb2pdbx-strip")
            else:
                dp.op("annot-rcsb2pdbx")

            logPath = os.path.join(self.__sessionPath, "annot-rcsb2pdbx.log")
            dp.expLog(logPath)
            dp.exp(outPath)
            if not self.__debug:
                dp.cleanup()
        except:  # noqa: E722 pylint: disable=bare-except
            traceback.print_exc(file=self.__lfh)
            return False
        #
        return True

    def rcsb2PdbxWithPdbId(self, inpPath, outPath):
        """RCSB CIF -> PDBx conversion  (converting to PDB ID entry/datablock id.)"""
        try:
            dp = RcsbDpUtility(tmpPath=self.__sessionPath, siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
            dp.imp(inpPath)
            dp.op("annot-rcsb2pdbx-withpdbid")
            logPath = os.path.join(self.__sessionPath, "annot-rcsb2pdbx.log")
            dp.expLog(logPath)
            dp.exp(outPath)
            if not self.__debug:
                dp.cleanup()
        except:  # noqa: E722 pylint: disable=bare-except
            traceback.print_exc(file=self.__lfh)
            return False
        #
        return True

    def rcsb2PdbxWithPdbIdAlt(self, inpPath, outPath):
        """RCSB CIF -> PDBx conversion  (converting to PDB ID entry/datablock id.)"""
        try:
            dp = RcsbDpUtility(tmpPath=self.__sessionPath, siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
            dp.imp(inpPath)
            dp.op("annot-rcsb2pdbx-alt")
            logPath = os.path.join(self.__sessionPath, "annot-rcsb2pdbxalt.log")
            dp.expLog(logPath)
            dp.exp(outPath)
            if not self.__debug:
                dp.cleanup()
        except:  # noqa: E722 pylint: disable=bare-except
            traceback.print_exc(file=self.__lfh)
            return False
        #
        return True

    def rcsbEps2Pdbx(self, inpPath, outPath, stripFlag=False, stripEntityFlag=False):
        """RCSB CIFEPS -> PDBx conversion (This still requires using the full maxit application)"""
        try:
            dp = RcsbDpUtility(tmpPath=self.__sessionPath, siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
            dp.imp(inpPath)
            if stripFlag:
                if stripEntityFlag:
                    dp.op("annot-rcsbeps2pdbx-strip-plus-entity")
                else:
                    dp.op("annot-rcsbeps2pdbx-strip")
            else:
                dp.op("annot-cif2cif")

            logPath = os.path.join(self.__sessionPath, "annot-rcsbeps2pdbx.log")
            dp.expLog(logPath)
            dp.exp(outPath)
            if not self.__debug:
                dp.cleanup()
        except:  # noqa: E722 pylint: disable=bare-except
            traceback.print_exc(file=self.__lfh)
            return False
        #
        return True

    def cif2Pdb(self, inpPath, outPath):
        """CIF -> PDB conversion  (Using the smaller application in the annotation package)"""
        try:
            dp = RcsbDpUtility(tmpPath=self.__sessionPath, siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
            dp.imp(inpPath)
            dp.op("annot-cif2pdb")
            logPath = os.path.join(self.__sessionPath, "annot-cif2pdb.log")
            dp.expLog(logPath)
            dp.exp(outPath)
            if not self.__debug:
                dp.cleanup()
        except:  # noqa: E722 pylint: disable=bare-except
            traceback.print_exc(file=self.__lfh)
            return False
        #
        return True

    def cif2Pdbx(self, inpPath, outPath):
        """CIF -> PDBx conversion  (public subset with PDBid conversion)"""
        try:
            dp = RcsbDpUtility(tmpPath=self.__sessionPath, siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
            dp.imp(inpPath)
            dp.op("cif2pdbx-public")
            logPath = os.path.join(self.__sessionPath, "annot-cif2pdbx.log")
            dp.expLog(logPath)
            dp.exp(outPath)
            if not self.__debug:
                dp.cleanup()
        except:  # noqa: E722 pylint: disable=bare-except
            traceback.print_exc(file=self.__lfh)
            return False
        #
        return True

    def modelConvertToPdbx(self, filePath=None, fileType="pdbx", pdbxFilePath=None):
        """Convert input model file format to PDBx.   Converted file is stored in the session
        directory using standard file naming.

        Return True for success or False otherwise.
        """
        if self.__verbose:
            self.__lfh.write("+DataFileAdapter.modelConvertToPdbx() filePath %s fileType %s pdbxFilePath %s\n" % (filePath, fileType, pdbxFilePath))
        try:
            ok = False
            if filePath is None or pdbxFilePath is None:
                return ok
            #
            if fileType in ["pdbx-mmcif", "pdbx", "pdbx-cif"]:
                if filePath != pdbxFilePath:
                    shutil.copyfile(filePath, pdbxFilePath)
                ok = True
            elif fileType == "rcsb-mmcif":
                ok = self.rcsb2Pdbx(filePath, pdbxFilePath, stripFlag=False)
            elif fileType == "rcsb-mmcif-strip":
                ok = self.rcsb2Pdbx(filePath, pdbxFilePath, stripFlag=True)
            elif fileType == "rcsb-cifeps":
                ok = self.rcsbEps2Pdbx(filePath, pdbxFilePath, stripFlag=False)
            elif fileType == "rcsb-cifeps-strip":
                ok = self.rcsbEps2Pdbx(filePath, pdbxFilePath, stripFlag=True)
                #
            else:
                ok = False
            return ok
        except:  # noqa: E722 pylint: disable=bare-except
            traceback.print_exc(file=self.__lfh)
            return False

    def pdbx2Assemblies(self, idCode, inpFilePath, outPath=".", indexFilePath=None):
        """Create model assemby files from input PDBx model file."""
        try:
            pdbxPath = inpFilePath
            logPath = os.path.join(self.__sessionPath, "pdbx-assembly.log")
            # indexFilePath=os.path.join(self.__sessionPath,"pdbx-assembly-index.txt")
            #
            dp = RcsbDpUtility(tmpPath=self.__sessionPath, siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
            dp.imp(pdbxPath)
            if idCode is not None:
                dp.addInput(name="deposition_data_set_id", value=idCode, type="param")

            if indexFilePath is not None:
                dp.addInput(name="index_file_path", value=indexFilePath, type="param")

            dp.op("annot-gen-assem-pdbx")
            dp.expLog(logPath)
            pthList = dp.getResultPathList()
            # wrkPath = dp.getWorkingDir()
            for pth in pthList:
                if os.access(pth, os.R_OK):
                    (_t, fn) = os.path.split(pth)
                    shutil.copyfile(pth, os.path.join(outPath, fn))

            # if (not self.__debug):
            #     dp.cleanup()
            #
            if self.__verbose:
                self.__lfh.write("+DataFileAdapter.pdbx2Assemblies() - input  model file path: %s\n" % inpFilePath)
                self.__lfh.write("+DataFileAdapter.pdbx2Assemblies() - assembly output paths:  %r\n" % pthList)
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            self.__lfh.write("+DataFileAdapter.pdbx2Assemblies() - failing for input file path %s output path %s\n" % (inpFilePath, outPath))
            traceback.print_exc(file=self.__lfh)
            return False

    def mtz2Pdbx(self, mtzFilePath, outSfFilePath, pdbxFilePath=None, logFilePath=None, diagsFilePath=None, dumpFilePath=None, timeout=120):  # pylint: disable=unused-argument
        """Convert input MTZ format to PDBx sf file."""
        try:
            diagfn = logFilePath if logFilePath is not None else "sf-convert-diags.cif"
            dmpfn = dumpFilePath if dumpFilePath is not None else "sf-convert-mtzdmp.log"
            #
            dp = RcsbDpUtility(tmpPath=self.__sessionPath, siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
            dp.imp(mtzFilePath)
            dp.setTimeout(timeout)
            if pdbxFilePath is not None:
                dp.addInput(name="xyz_file_path", value=pdbxFilePath)
            dp.op("annot-sf-convert")
            dp.expLog(logFilePath)
            dp.expList(dstPathList=[outSfFilePath, diagfn, dmpfn])

            if not self.__debug:
                dp.cleanup()
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            self.__lfh.write("+DataFileAdapter.mtz2Pdbx() - failing for mtz file path %s output path %s\n" % (mtzFilePath, outSfFilePath))
            traceback.print_exc(file=self.__lfh)
            return False



##
# File:    PdbxMergeCategory.py
##
"""  Merge/replace selected categories from the first data container and
     write the result.
"""

__docformat__ = "restructuredtext en"
__author__ = "Ezra Peisach"
__email__ = "ezra.peisach@rcsb.org"
__license__ = "Apache 2.0"


import logging

from mmcif.io.IoAdapterCore import IoAdapterCore

logger = logging.getLogger(__name__)


class PdbxMergeCategory(object):
    def __init__(self):
        pass

    @staticmethod
    def merge(srcpath, newcontentpath, outpath, mergelist=None, replacelist=None):
        """Merges selected categories from newcontentpath (first block) into srcpath and outputs to outpath in
                   first data block.
        Categories in replacelist will replace category, if present in newcontentpath, otherwise leaves alone.
        Categories in mergelist must be single row. Attributes will be replaced/appended. Attributes not in
                   newcontentpath are ignored.

        newcontentpath will contain to combined file

        If the same category is in mergelist and replacelist, undefined behaviour

        """
        logger.debug("Starting merge %s %s %s", srcpath, newcontentpath, outpath)
        try:
            io = IoAdapterCore()
            srcin = io.readFile(srcpath)
            srcblock = srcin[0]

            newcontent = io.readFile(newcontentpath)
            newcontentblock = newcontent[0]

            # The plan is to replace/modify the datablock - and then rewrite
            if replacelist:
                for cat in replacelist:
                    if cat in newcontentblock.getObjNameList():
                        obj = newcontentblock.getObj(cat)
                        srcblock.append(obj)

            if mergelist:
                for cat in mergelist:
                    if cat in newcontentblock.getObjNameList():
                        obj = newcontentblock.getObj(cat)
                        if srcblock.exists(cat):
                            # Category exists in file
                            srcobj = srcblock.getObj(cat)
                            for attr in obj.getAttributeList():
                                val = obj.getValue(attr, 0)
                                if not srcobj.hasAttribute(attr):
                                    srcobj.appendAttribute(attr)
                                srcobj.setValue(val, attr, 0)
                        else:
                            # New category
                            srcblock.append(obj)

            ret = io.writeFile(outpath, srcin)
            return ret

        except Exception as e:
            logger.exception("Failing with %s", str(e))
            return False



#!/usr/bin/env python3

import argparse
import logging
import os
import sys

from wwpdb.utils.dp.electron_density.common_functions import convert_mdb_to_binary_cif, run_command_and_check_output_file

logger = logging.getLogger()


class EmVolumes:
    def __init__(
        self,
        em_map,
        node_path,
        volume_server_pack_path,
        volume_server_query_path,
        binary_map_out,
        working_dir,
    ):

        self.em_map = em_map
        self.em_map_name = os.path.basename(em_map)
        self.mdb_map = "em_map.mdb"
        self.node_path = node_path
        self.volume_server_pack_path = volume_server_pack_path
        self.volume_server_query_path = volume_server_query_path
        self.mdb_map_path = None
        self.bcif_map_path = binary_map_out
        self.workdir = working_dir if working_dir else os.getcwd()

    def run_conversion(self):
        bcif_dir_out = os.path.dirname(self.bcif_map_path)
        if bcif_dir_out:
            if not os.path.exists(bcif_dir_out):
                os.makedirs(bcif_dir_out)
        logging.debug("temp working folder: %s", self.workdir)
        self.mdb_map_path = os.path.join(self.workdir, self.mdb_map)

        worked = self.make_volume_server_map()

        if worked:
            worked = self.convert_map_to_binary_cif()

        return worked

    def make_volume_server_map(self):
        if os.path.exists(self.em_map):
            command = "%s %s em %s %s" % (self.node_path, self.volume_server_pack_path, self.em_map, self.mdb_map_path)
            logging.debug(command)
            return run_command_and_check_output_file(command=command, process_name="make Volume server map", workdir=self.workdir, output_file=self.mdb_map_path)
        else:
            logging.error("input map file missing: %s", self.em_map)
        return False

    def convert_map_to_binary_cif(self):
        return convert_mdb_to_binary_cif(
            node_path=self.node_path,
            volume_server_query_path=self.volume_server_query_path,
            map_id="em_volume",
            source_id="em",
            output_file=self.bcif_map_path,
            working_dir=self.workdir,
            mdb_map_path=self.mdb_map_path,
            detail=1,
        )


def main():  # pragma: no cover
    parser = argparse.ArgumentParser()
    parser.add_argument("--em_map", help="EM map", type=str, required=True)
    parser.add_argument("--working_dir", help="working dir", type=str, required=True)
    parser.add_argument("--binary_map_out", help="Output filename of binary map", type=str, required=True)
    parser.add_argument("--node_path", help="path to node", type=str, required=True)
    parser.add_argument("--volume_server_pack_path", help="path to volume-server-pack", type=str, required=True)
    parser.add_argument("--volume_server_query_path", help="path to volume-server-query", type=str, required=True)
    parser.add_argument("--keep_working_directory", help="keep working directory", action="store_true")
    parser.add_argument("--debug", help="debugging", action="store_const", dest="loglevel", const=logging.DEBUG, default=logging.INFO)

    args = parser.parse_args()
    logger.setLevel(args.loglevel)

    if not sys.argv[1:]:
        parser.print_help()
        exit()

    em = EmVolumes(
        em_map=args.em_map,
        node_path=args.node_path,
        volume_server_pack_path=args.volume_server_pack_path,
        volume_server_query_path=args.volume_server_query_path,
        binary_map_out=args.binary_map_out,
        working_dir=args.working_dir,
    )
    worked = em.run_conversion()
    logging.info("EM map conversion worked: {}".format(worked))  # pylint: disable=logging-format-interpolation
    if not worked:
        sys.exit(1)


if __name__ == "__main__":
    main()
